@@ -1,109 +1,98 @@
 import atexit
-import inspect
 
 from giving import SourceProxy
 
-from .core import dict_to_pattern_list, global_patterns
-from .deco import tooled
+from .interpret import Immediate, Total
+from .overlay import BaseOverlay, autotool
 from .selector import select
-from .selfless import override
-from .tags import tag
 from .utils import ABSENT
 
 global_probes = set()
 
 
-def make_resolver(*namespaces):
-    """Hook into ptera's selector resolution.
-
-    * Resolve functions in the given namespaces. They will automatically be
-      instrumented with ptera.tooled.
-    * When a tag is found, instrument all functions in all modules that might
-      use that tag (with ptera.tooled).
-
-    Arguments:
-        namespaces: A list of globals dicts to find functions and variables in.
-    """
-
-    def __ptera_resolver__(x):
-        if x.startswith("/"):
-            import codefind
-
-            _, module, *parts = x.split("/")
-            co = codefind.find_code(*parts, module=module or "__main__")
-            funcs = [
-                fn
-                for fn in codefind.get_functions(co)
-                if inspect.isfunction(fn)
-                and not fn.__name__.endswith("__ptera_redirect")
-            ]
-            (curr,) = funcs
-
-        else:
-            varname, *rest = x.split(".")
-
-            if varname.startswith("@"):
-                curr = getattr(tag, varname[1:])
-
-            else:
-                for ns in namespaces:
-                    if varname in ns:
-                        curr = ns[varname]
-                        seq = [(ns, varname, dict.__setitem__)]
-                        break
-                else:
-                    raise NameError(f"Could not resolve '{varname}'.")
-
-                for part in rest:
-                    seq.append((curr, part, setattr))
-                    curr = getattr(curr, part)
-
-        if inspect.isfunction(curr):
-            # Instrument the function directly
-            tooled.inplace(curr)
-
-        # elif isinstance(curr, Tag):
-        #     # Instrument existing modules and update substitutions
-        #     instrument_for_tag(curr)
-
-        return getattr(curr, "__ptera__", curr)
-
-    return __ptera_resolver__
+def _identity(x):
+    return x
 
 
 class Probe(SourceProxy):
     """Observable which generates a stream of values from program variables.
 
-    Probes should be created with `ptera.probing` or `ptera.global_probe`.
+    Probes should be created with
+    :func:`~ptera.probe.probing` or :func:`~ptera.probe.global_probe`.
+
+    .. note::
+        In the documentation for some methods you may see calls to ``give()``
+        or ``given()``, but that's because they come from the documentation
+        for the ``giving`` package (on top of which Probe is built).
+
+        ``give()`` is equivalent to what Ptera does when a
+        variable of interest is set, ``given()`` yields an object that has
+        the same interface as ``Probe`` (the superclass to ``Probe``, in
+        fact). Take variables named ``gv`` to be probes.
 
     Arguments:
         selectors: The selector strings describing the variables to probe (at least one).
-        raw: Defaults to False. If True, produce a stream of Capture objects that
+        raw: Defaults to False. If True, produce a stream of :class:`~ptera.interpret.Capture` objects that
             contain extra information about the capture. Mostly relevant for
-            advanced selectors such as "f > $x:@Parameter" which captures the value
+            advanced selectors such as ``f > $x:@Parameter`` which captures the value
             of any variable with the Parameter tag under the generic name "x".
             When raw is True, the actual name of the variable is preserved in a
             Capture object associated to x.
+        probe_type: Either "immediate", "total", or None (the default).
+
+            * If "immediate", use :class:`~ptera.interpret.Immediate`.
+            * If "total", use :class:`~ptera.interpret.Total`.
+            * If None, determine what to use based on whether the selector has
+              a focus or not.
+        env: A dictionary that will be used to resolve symbols in the selector.
+            If it is not provided, ptera will seek the locals and globals
+            dictionaries of the scope where this function is called.
     """
 
-    def __init__(self, *selectors, raw=False, _obs=None, _root=None):
+    def __init__(
+        self,
+        *selectors,
+        raw=False,
+        probe_type=None,
+        env=None,
+        _obs=None,
+        _root=None
+    ):
+        # Note: the private _obs and _root parameters are used to "fork"
+        # the probe when using operators on it while keeping a reference
+        # to the root or master probe.
+
         if not selectors and _obs is None:
             raise TypeError("Probe() takes at least one selector argument.")
 
         super().__init__(_obs=_obs, _root=_root)
 
         if selectors:
-            self._selectors = [
-                select(selector, env_wrapper=make_resolver)
-                for selector in selectors
+            if probe_type not in ("immediate", "total", None):
+                raise TypeError(
+                    "probe_type must be 'immediate', 'total' or None"
+                )
+            self._selectors = [select(s, env=env) for s in selectors]
+            rules = [
+                Immediate(sel, intercept=self._emit)
+                if probe_type != "total"
+                and (sel.focus or probe_type == "immediate")
+                else Total(sel, close=self._emit)
+                for sel in self._selectors
             ]
-            self._patterns = dict_to_pattern_list(
-                {sel: {"value": self._emit} for sel in self._selectors}
-            )
+            self._ol = BaseOverlay(*rules)
             self._raw = raw
             self._activated = False
 
-    def override(self, setter=lambda x: x):
+    def _install_tooling(self):
+        for selector in self._selectors:
+            autotool(selector)
+
+    def _uninstall_tooling(self):
+        for selector in self._selectors:
+            autotool(selector, undo=True)
+
+    def override(self, setter=_identity):
         """Override the value of the focus variable using a setter function.
 
         .. code-block:: python
@@ -135,7 +124,7 @@ class Probe(SourceProxy):
                 A function that takes a value from the pipeline and produces the value
                 to set the focus variable to.
 
-                * If not set, the value from the stream is used as-is.
+                * If not provided, the value from the stream is used as-is.
                 * If not callable, set the variable to the value of setter
         """
         if not callable(setter):
@@ -145,7 +134,7 @@ class Probe(SourceProxy):
                 return value
 
         def _override(data):
-            self._root._value = override(setter(data))
+            self._root._value = setter(data)
 
         return self.subscribe(_override)
 
@@ -167,7 +156,7 @@ class Probe(SourceProxy):
             **Important:** override() only overrides the **focus variable**. The focus
             variable is the one to the right of ``>``, or the one prefixed with ``!``.
 
-            See :meth:`~ptera.probe.ProbeProxy.override`.
+            See :meth:`~ptera.probe.Probe.override`.
 
         Arguments:
             setter: A function that takes a value from the pipeline as keyword arguments
@@ -175,7 +164,7 @@ class Probe(SourceProxy):
         """
 
         def _override(data):
-            self._root._value = override(setter(**data))
+            self._root._value = setter(**data)
 
         return self.subscribe(_override)
 
@@ -199,7 +188,7 @@ class Probe(SourceProxy):
     # Context manager #
     ###################
 
-    def _emit(self, **data):
+    def _emit(self, data):
         """Emit data on the stream.
 
         This is used internally.
@@ -213,20 +202,35 @@ class Probe(SourceProxy):
         return self._value
 
     def _enter(self):
+        # This is called on the root probe by the __enter__ method of a child.
+        # So e.g. `with probing("f > x").min(): ...` will call __enter__ on the
+        # object returned by min, but _enter is called on the main probe
+        # returned by probing (stored in self._root)
         if self._activated:
             raise Exception("An instance of Probe can only be entered once")
 
+        self._install_tooling()
         self._activated = True
         global_probes.add(self)
-        global_patterns.extend(self._patterns)
+        self._ol.__enter__()
         return self
 
     def _exit(self):
-        global_patterns.remove_all(self._patterns)
+        # This is called on the root probe by the __exit__ method of a child.
+        self._ol.__exit__(None, None, None)
         global_probes.remove(self)
+        self._uninstall_tooling()
+
+    def activate(self):
+        """Activate this probe."""
+        self.__enter__()
+
+    def deactivate(self):
+        """Deactivate this probe."""
+        self.__exit__(None, None, None)
 
 
-def probing(*selectors, raw=False):
+def probing(*selectors, raw=False, probe_type=None, env=None):
     """Probe that can be used as a context manager.
 
     Example:
@@ -240,15 +244,22 @@ def probing(*selectors, raw=False):
 
     Arguments:
         selectors: The selector strings describing the variables to probe (at least one).
-        do: A function to execute on each data point.
-        format: A format string (implies do=print)
-        raw: Defaults to False. If True, produce a stream of Capture objects that
+        raw: Defaults to False. If True, produce a stream of :class:`~ptera.interpret.Capture` objects that
             contain extra information about the capture.
+        probe_type: Either "immediate", "total", or None (the default).
+
+            * If "immediate", use :class:`~ptera.interpret.Immediate`.
+            * If "total", use :class:`~ptera.interpret.Total`.
+            * If None, determine what to use based on whether the selector has
+              a focus or not.
+        env: A dictionary that will be used to resolve symbols in the selector.
+            If it is not provided, ptera will seek the locals and globals
+            dictionaries of the scope where this function is called.
     """
-    return Probe(*selectors, raw=raw)
+    return Probe(*selectors, raw=raw, probe_type=probe_type, env=env)
 
 
-def global_probe(*selectors, raw=False):
+def global_probe(*selectors, raw=False, probe_type=None, env=None):
     """Set a probe globally.
 
     Example:
@@ -263,15 +274,27 @@ def global_probe(*selectors, raw=False):
 
     Arguments:
         selectors: The selector strings describing the variables to probe (at least one).
-        raw: Defaults to False. If True, produce a stream of Capture objects that
+        raw: Defaults to False. If True, produce a stream of :class:`~ptera.interpret.Capture` objects that
             contain extra information about the capture.
+        probe_type: Either "immediate", "total", or None (the default).
+
+            * If "immediate", use :class:`~ptera.interpret.Immediate`.
+            * If "total", use :class:`~ptera.interpret.Total`.
+            * If None, determine what to use based on whether the selector has
+              a focus or not.
+        env: A dictionary that will be used to resolve symbols in the selector.
+            If it is not provided, ptera will seek the locals and globals
+            dictionaries of the scope where this function is called.
     """
-    prb = Probe(*selectors, raw=raw)
-    prb.__enter__()
+    prb = Probe(*selectors, raw=raw, probe_type=probe_type, env=env)
+    prb.activate()
     return prb
 
 
 @atexit.register
 def _terminate_global_probes():  # pragma: no cover
+    # This closes active probes at program exit. This is important for global
+    # probes if reduction operations like min() are requested, because it tells
+    # them that there is no more data and that they can proceed.
     for probe in list(global_probes):
-        probe.__exit__(None, None, None)
+        probe.deactivate()
