@@ -1,14 +1,38 @@
 """Specifications for call paths."""
 
 
-import builtins
+import inspect
 import re
 import sys
+import types
+from collections import defaultdict
 from itertools import count
 
 from . import opparse
-from .tags import Tag, tag as tag_factory
-from .utils import ABSENT
+from .tags import Tag, match_tag, tag as tag_factory
+from .utils import (
+    ABSENT,
+    CodeNotFoundError,
+    DictPile,
+    cached_property,
+    is_tooled,
+)
+
+_valid_hashvars = ("#enter", "#value", "#yield")
+
+
+class SelectorError(Exception):
+    """Error raised for invalid selectors."""
+
+
+def check_element(el, name, category):
+    """Check if Element el matches the given name and category."""
+    if el.name is not None and el.name != name:
+        return False
+    elif not match_tag(el.category, category):
+        return False
+    else:
+        return True
 
 
 class InternedMC(type):
@@ -24,7 +48,28 @@ class InternedMC(type):
         return cls._cache[key]
 
 
-class Element(metaclass=InternedMC):
+class Selector(metaclass=InternedMC):
+    """Represents a selector for variables in a call stack."""
+
+    def check_captures(self, captures):
+        for v in self.all_values:
+            if v.capture in captures:
+                cap = captures[v.capture]
+                for value in cap.values:
+                    match = v.value == value or (
+                        isinstance(v.value, MatchFunction) and v.value.fn(value)
+                    )
+                    if not match:
+                        return False
+        return True
+
+    def __str__(self):
+        return f'sel("{self.encode()}")'
+
+    __repr__ = __str__
+
+
+class Element(Selector):
     """Represents a variable or some other atom."""
 
     _constructor_defaults = {
@@ -32,18 +77,45 @@ class Element(metaclass=InternedMC):
         "category": None,
         "capture": None,
         "tags": frozenset(),
-        "key_field": None,
     }
 
-    def __init__(self, *, name, value, category, capture, tags, key_field):
+    def __init__(self, *, name, value, category, capture, tags):
         self.name = name
         self.value = value
         self.category = category
         self.capture = capture
         self.tags = tags
-        self.key_field = key_field
-        self.focus = 1 in self.tags
-        self.hasval = self.value is not ABSENT
+
+    @cached_property
+    def focus(self):
+        return 1 in self.tags
+
+    @cached_property
+    def hasval(self):
+        return self.value is not ABSENT
+
+    @cached_property
+    def main(self):
+        return self if self.focus else None
+
+    @cached_property
+    def all_captures(self):
+        if self.capture and not self.capture.startswith("/"):
+            return {self.capture}
+        else:  # pragma: no cover
+            # Does not currently happen
+            return set()
+
+    @cached_property
+    def all_values(self):
+        return [self] if self.hasval else []
+
+    @cached_property
+    def valid(self):
+        if self.name is None:
+            return self.focus
+        else:
+            return True
 
     def with_focus(self):
         return self.clone(tags=self.tags | frozenset({1}))
@@ -58,44 +130,22 @@ class Element(metaclass=InternedMC):
             "category": self.category,
             "capture": self.capture,
             "tags": self.tags,
-            "key_field": self.key_field,
             **changes,
         }
         return Element(**args)
 
-    def all_captures(self):
-        if self.capture and not self.capture.startswith("/"):
-            return {self.capture}
-        else:
-            return set()
-
-    def valid(self):
-        if self.name is None:
-            return self.focus
-        else:
-            return True
-
     def rewrite(self, required, focus=None):
         if focus is not None and focus == self.capture:
             return self.with_focus()
         elif focus is None and self.focus:
             return self
         elif self.capture not in required:
-            if self.value is ABSENT:
-                return None
-            else:
-                return self.clone(capture=None).without_focus()
+            return None
         elif focus is not None:
             return self.without_focus()
         else:
             return self
 
-    def key_captures(self):
-        if self.key_field is not None:
-            return {(self.capture, self.key_field)}
-        else:
-            return set()
-
     def specialize(self, specializations):
         """Replace $variables in the selector using a specializations dict."""
         spc = specializations.get(self.capture, ABSENT)
@@ -106,10 +156,6 @@ class Element(metaclass=InternedMC):
             category=self.category if spc.category is None else spc.category,
             value=self.value if spc.value is ABSENT else spc.value,
         )
-        if rval.key_field == "name" and rval.name is not None:
-            rval = rval.clone(key_field=None)
-        if rval.key_field == "value" and rval.value is not ABSENT:
-            rval = rval.clone(key_field=None)
         return rval
 
     def encode(self):
@@ -129,30 +175,68 @@ class Element(metaclass=InternedMC):
         val = f"={self.value}" if self.value is not ABSENT else ""
         return f"{focus}{name}{cap}{cat}{val}"
 
-    def __str__(self):
-        return f'sel("{self.encode()}")'
 
-    __repr__ = __str__
-
-
-class Call(metaclass=InternedMC):
+class Call(Selector):
     """Represents a call in the call stack."""
 
     _constructor_defaults = {
         "children": (),
         "captures": (),
         "immediate": False,
-        "collapse": False,
     }
 
-    def __init__(self, *, element, children, captures, immediate, collapse):
+    def __init__(self, *, element, children, captures, immediate):
         self.element = element
         self.children = children
         self.captures = captures
         self.immediate = immediate
-        self.collapse = collapse
-        self.focus = any(x.focus for x in self.captures + self.children)
-        self.hasval = any(x.hasval for x in self.captures + self.children)
+
+    @cached_property
+    def focus(self):
+        return any(x.focus for x in self.captures + self.children)
+
+    @cached_property
+    def hasval(self):
+        return any(x.hasval for x in self.captures + self.children)
+
+    @cached_property
+    def all_tags(self):
+        results = defaultdict(set)
+        for child in self.children:
+            for tag, values in child.all_tags.items():
+                results[tag] |= values
+        for cap in self.captures:
+            for tag in cap.tags:
+                results[tag].add(cap)
+        return results
+
+    @cached_property
+    def main(self):
+        for x in self.captures + self.children:
+            if x.main is not None:
+                return x.main
+        return None
+
+    @cached_property
+    def all_captures(self):
+        rval = set()
+        for x in self.captures + self.children:
+            rval.update(x.all_captures)
+        return rval
+
+    @cached_property
+    def all_values(self):
+        rval = []
+        for x in self.captures + self.children:
+            rval += x.all_values
+        return rval
+
+    @cached_property
+    def valid(self):
+        return (
+            all(x.valid for x in self.captures + self.children)
+            and sum(x.focus for x in self.captures + self.children) <= 1
+        )
 
     def clone(self, **changes):
         args = {
@@ -160,32 +244,10 @@ class Call(metaclass=InternedMC):
             "children": self.children,
             "captures": self.captures,
             "immediate": self.immediate,
-            "collapse": self.collapse,
             **changes,
         }
         return Call(**args)
 
-    def find_tag(self, tag):
-        results = set()
-        for child in self.children:
-            results |= child.find_tag(tag)
-        for cap in self.captures:
-            if tag in cap.tags:
-                results.add(cap)
-        return results
-
-    def all_captures(self):
-        rval = set()
-        for x in self.captures + self.children:
-            rval.update(x.all_captures())
-        return rval
-
-    def valid(self):
-        return (
-            all(x.valid() for x in self.captures + self.children)
-            and sum(x.focus for x in self.captures + self.children) <= 1
-        )
-
     def rewrite(self, required, focus=None):
         captures = [x.rewrite(required, focus) for x in self.captures]
         captures = [x for x in captures if x is not None]
@@ -198,14 +260,6 @@ class Call(metaclass=InternedMC):
 
         return self.clone(captures=tuple(captures), children=tuple(children))
 
-    def key_captures(self):
-        rval = self.element.key_captures()
-        for child in self.children:
-            rval.update(child.key_captures())
-        for cap in self.captures:
-            rval.update(cap.key_captures())
-        return rval
-
     def specialize(self, specializations):
         """Replace $variables in the selector using a specializations dict."""
         return self.clone(
@@ -218,6 +272,17 @@ class Call(metaclass=InternedMC):
             ),
         )
 
+    def wrap_functions(self, wrap):
+        element = self.element.clone(
+            name=wrap(self.element.name, self.captures)
+        )
+        return self.clone(
+            element=element,
+            children=tuple(
+                child.wrap_functions(wrap) for child in self.children
+            ),
+        )
+
     def encode(self):
         """Return a string representation of the selector."""
         name = self.element.encode()
@@ -225,18 +290,67 @@ class Call(metaclass=InternedMC):
         for cap in self.captures:
             caps.append(cap.encode())
         for child in self.children:
-            enc = child.encode()
-            enc = f"> {enc}" if child.immediate else f">> {enc}"
-            caps.append(enc)
+            caps.append(child.encode())
         caps = "" if not caps else "(" + ", ".join(caps) + ")"
         return f"{name}{caps}"
 
-    def __str__(self):
-        return f'sel("{self.encode()}")'
+    def problems(self):
+        """Return a list of problems with this selector.
 
-    __repr__ = __str__
+        * Wildcards are not allowed for cuntions.
+        * All functions should be tooled.
+        * All captured variables should exist in their respective functions.
+        * For wildcard variables that specify a tag/category, at least one
+          variable should match.
+        """
+        problems = []
+
+        func = self.element.name
+        info = getattr(func, "__ptera_info__", None)
+
+        if func is None:
+            problems.append("Wildcard function is not allowed")
+
+        elif info is None:
+            problems.append(f"{func} is not properly tooled")
+
+        else:
+            for x in self.captures:
+                if x.name is None:
+                    for _, data in info.items():
+                        if check_element(x, x.name, data["annotation"]):
+                            break
+                    else:
+                        problems.append(
+                            f"No variable in `{func}` has the category `{x.category}`"
+                        )
+
+                elif x.name.startswith("#"):
+                    if x.name not in _valid_hashvars:
+                        problems.append(
+                            f"{x.name} is not a valid hashvar. Valid hashvars are: {', '.join(_valid_hashvars)}"
+                        )
+
+                else:
+                    name = x.name.split(".")[0]
+                    data = info.get(name, None)
+                    if not data:
+                        problems.append(
+                            f"Cannot find a variable named `{x.name}` in `{func}`"
+                        )
+
+                    elif not check_element(x, x.name, data["annotation"]):
+                        problems.append(
+                            f"Variable `{func} > {x.name}` does not have the category `{x.category}`"
+                        )
+
+        for x in self.children:
+            problems.extend(x.problems())
+
+        return problems
 
 
+# Parser for selectors
 parser = opparse.Parser(
     lexer=opparse.Lexer(
         {
@@ -273,13 +387,14 @@ def _guarantee_call(parent, context, resolve=True):
     if isinstance(parent, Element):
         name = VSymbol(parent.name) if parent.name and resolve else parent.name
         parent = parent.clone(capture=None, name=name).without_focus()
-        immediate = context == "incall"
-        parent = Call(element=parent, captures=(), immediate=immediate)
+        parent = Call(element=parent, captures=(), immediate=False)
     assert isinstance(parent, Call)
     return parent
 
 
 class Evaluator:
+    """Evaluator that transforms the parse tree into a Selector."""
+
     def __init__(self):
         self.actions = {}
 
@@ -324,49 +439,25 @@ def make_nested_imm(node, parent, child, context):
         return parent.clone(captures=parent.captures + (child,))
     else:
         return parent.clone(
-            children=parent.children + (child.clone(immediate=True),),
+            children=parent.children + (child.clone(immediate=False),),
         )
 
 
-@evaluate.register_action("X >> X")
-def make_nested(node, parent, child, context):
-    parent = evaluate(parent, context=context)
-    child = evaluate(child, context=context)
-    parent = _guarantee_call(parent, context=context)
-    if isinstance(child, Element):
-        child = child.with_focus()
-        child = Call(
-            element=Element(name=None),
-            captures=(child,),
-            immediate=False,
-            collapse=True,
-        )
-    return parent.clone(children=parent.children + (child,))
-
-
-@evaluate.register_action("_ > X")
-def make_nested_imm_pfx(node, _, child, context):
-    child = evaluate(child, context=context)
-    if isinstance(child, Element):
-        return Call(
-            element=Element(name=None), captures=(child,), immediate=True
-        )
-    else:
-        return child.clone(immediate=True)
-
-
-@evaluate.register_action("_ >> X")
-def make_nested_pfx(node, _, child, context):
-    child = evaluate(child, context=context)
-    if isinstance(child, Element):
-        return Call(
-            element=Element(name=None),
-            captures=(child,),
-            immediate=False,
-            collapse=True,
-        )
-    else:
-        return child.clone(immediate=False)
+# LEGACY
+# @evaluate.register_action("X >> X")
+# def make_nested(node, parent, child, context):
+#     parent = evaluate(parent, context=context)
+#     child = evaluate(child, context=context)
+#     parent = _guarantee_call(parent, context=context)
+#     if isinstance(child, Element):
+#         child = child.with_focus()
+#         child = Call(
+#             element=Element(name=None),
+#             captures=(child,),
+#             immediate=False,
+#             collapse=True,
+#         )
+#     return parent.clone(children=parent.children + (child,))
 
 
 @evaluate.register_action("_ : X")
@@ -396,44 +487,7 @@ def make_double_focus(node, _, element, context):
 @evaluate.register_action("_ $ X")
 def make_dollar(node, _, name, context):
     name = evaluate(name, context=context)
-    return Element(
-        name=None, category=None, capture=name.name, key_field="name"
-    )
-
-
-@evaluate.register_action("X [ X ] _")
-def make_index(node, element, key, _, context, resolve_call=False):
-    def _make_key(key, suffix):
-        assert isinstance(key, Element)
-        if key.value is not ABSENT:
-            assert key.name is None
-            val = key.value
-        else:
-            val = VSymbol(key.name) if key.name is not None else ABSENT
-        return Element(
-            name=f"#key{suffix}",
-            value=val,
-            category=key.category,
-            capture=key.capture if key.name != key.capture else None,
-            key_field="value" if key.name is None else None,
-        )
-
-    element = evaluate(element, context=context)
-    key = evaluate(key, context=context)
-    assert isinstance(element, Element)
-    element = _guarantee_call(element, context=context, resolve=resolve_call)
-
-    if isinstance(key, list):
-        keys = tuple(_make_key(key, i) for i, key in enumerate(key))
-    else:
-        keys = (_make_key(key, ""),)
-
-    return element.clone(captures=element.captures + keys)
-
-
-@evaluate.register_action("X [[ X ]] _")
-def make_function_index(node, element, key, _, context):
-    return make_index(node, element, key, _, context, resolve_call=True)
+    return Element(name=None, category=None, capture=name.name)
 
 
 @evaluate.register_action("X ( _ ) _")
@@ -464,10 +518,7 @@ def make_as(node, element, name, context):
     element = evaluate(element, context=context)
     name = evaluate(name, context=context)
     if isinstance(element, Element):
-        return element.clone(
-            capture=name.name,
-            key_field="name" if element.name is None else None,
-        )
+        return element.clone(capture=name.name)
     else:
         focus = context == "root"
         new_capture = Element(
@@ -478,25 +529,19 @@ def make_as(node, element, name, context):
         return element.clone(captures=element.captures + (new_capture,))
 
 
-@evaluate.register_action("_ = X")
 @evaluate.register_action("X = X")
 def make_equals(node, element, value, context, matchfn=False):
-    if element is None:
-        element = Element(name=None)
-    else:
-        element = evaluate(element, context=context)
+    element = evaluate(element, context=context)
     value = value_evaluate(value)
     if matchfn:
         value = VCall(MatchFunction, (value,))
     if isinstance(element, Element):
-        capture = element.capture if matchfn else None
-        return element.clone(value=value, capture=capture)
+        return element.clone(value=value, capture=element.capture)
     else:
-        new_element = Element(name="#value", value=value, capture=None)
+        new_element = Element(name="#value", value=value, capture="#value")
         return element.clone(captures=element.captures + (new_element,))
 
 
-@evaluate.register_action("_ ~ X")
 @evaluate.register_action("X ~ X")
 def make_matchfn(node, element, value, context):
     return make_equals(node, element, value, context, matchfn=True)
@@ -507,12 +552,10 @@ def make_symbol(node, context):
     if node.value == "*":
         element = Element(name=None)
     else:
-        value = node.value
-        cap = node.value[1:] if node.value.startswith("#") else node.value
         focus = context == "root"
         element = Element(
-            name=value,
-            capture=cap,
+            name=node.value,
+            capture=node.value,
             tags=frozenset({1}) if focus else frozenset(),
         )
     return element
@@ -522,19 +565,50 @@ def dict_resolver(env):
     """Resolve a symbol from a dictionary, e.g. the globals directory."""
 
     def resolve(x):
-        if x.startswith("@"):
+        if x.startswith("/"):
+            import codefind
+
+            _, module, *hierarchy = x.split("/")
+            if any("." in part for part in hierarchy):
+                raise SelectorError(
+                    "Only the module part of a /selector can contain dots."
+                    " Try calling `ptera.refstring` on the function you want"
+                    " to select. It will return the proper way to refer to it."
+                )
+
+            try:
+                co = codefind.find_code(*hierarchy, module=module or "__main__")
+            except KeyError:
+                raise CodeNotFoundError(
+                    f"Cannot find a function for the reference '{x}'."
+                    " Try calling `ptera.refstring` on the function you want"
+                    " to select. It will return the proper way to refer to it."
+                )
+
+            funcs = [
+                fn
+                for fn in codefind.get_functions(co)
+                if inspect.isfunction(fn)
+                and not getattr(fn, "__ptera_discard__", False)
+            ]
+            if not funcs:  # pragma: no cover
+                raise Exception(f"Reference `{x}` cannot be resolved.")
+            elif len(funcs) > 1:  # pragma: no cover
+                raise Exception(f"Reference `{x}` is ambiguous.")
+            (curr,) = funcs
+
+        elif x.startswith("@"):
             return getattr(tag_factory, x[1:])
 
-        start, *parts = x.split(".")
-        if start in env:
-            curr = env[start]
-        elif hasattr(builtins, start):
-            return getattr(builtins, start)
         else:
-            raise Exception(f"Could not resolve '{start}'.")
+            start, *parts = x.split(".")
+            if start in env:
+                curr = env[start]
+            else:
+                raise SelectorError(f"Could not resolve '{start}'.")
 
-        for part in parts:
-            curr = getattr(curr, part)
+            for part in parts:
+                curr = getattr(curr, part)
 
         return getattr(curr, "__ptera__", curr)
 
@@ -558,7 +632,7 @@ class VSymbol(VNode):
             return int(x)
         elif re.fullmatch(r"'[^']*'", x):
             return x[1:-1]
-        elif isinstance(env, dict):
+        elif isinstance(env, (dict, DictPile)):
             return dict_resolver(env)(x)
         else:
             return env(x)
@@ -669,7 +743,7 @@ def _find_eval_env(s, fr, skip):
             return glb["__ptera_resolver__"]
         name = glb["__name__"]
         if all(not name.startswith(pfx) for pfx in skip):
-            return glb
+            return DictPile(fr.f_locals, glb, __builtins__)
         fr = fr.f_back
     raise AssertionError("Unreachable outside ptera.")  # pragma: no cover
 
@@ -679,42 +753,67 @@ class MatchFunction:
         self.fn = fn
 
 
-def _resolve(pattern, env, cnt):
-    if isinstance(pattern, Call):
-        el = _resolve(pattern.element, env, cnt)
-        return pattern.clone(
+def _dig(fn):
+    while hasattr(fn, "__wrapped__") and not is_tooled(fn):
+        fn = fn.__wrapped__
+    return fn
+
+
+def _resolve(selector, env, cnt):
+    if isinstance(selector, Call):
+        el = _resolve(selector.element, env, cnt)
+        captures = [_resolve(x, env, cnt) for x in selector.captures]
+        fn = el.name
+        if isinstance(fn, types.MethodType):
+            # If fn is a method, we add a capture for "self" that must
+            # match the instance.
+            real_fn = _dig(fn.__func__)
+            selfname = inspect.getfullargspec(real_fn).args[0]
+            el = el.clone(name=real_fn)
+            captures.append(
+                Element(
+                    name=selfname,
+                    capture=selfname,
+                    value=fn.__self__,
+                )
+            )
+        else:
+            unwrapped = _dig(fn)
+            if fn is not unwrapped:
+                el = el.clone(name=unwrapped)
+        return selector.clone(
             element=el,
-            captures=tuple(_resolve(x, env, cnt) for x in pattern.captures),
-            children=tuple(_resolve(x, env, cnt) for x in pattern.children),
+            captures=tuple(captures),
+            children=tuple(_resolve(x, env, cnt) for x in selector.children),
         )
-    elif isinstance(pattern, Element):
-        name = _eval(pattern.name, env)
-        category = _eval(pattern.category, env)
-        value = _eval(pattern.value, env)
+    elif isinstance(selector, Element):
+        name = _eval(selector.name, env)
+        category = _eval(selector.category, env)
+        value = _eval(selector.value, env)
         capture = (
-            f"/{next(cnt)}" if pattern.capture is None else pattern.capture
+            f"/{next(cnt)}" if selector.capture is None else selector.capture
         )
         if category is not None and not isinstance(category, Tag):
-            raise TypeError("A pattern can only be a Tag.")
-        return pattern.clone(
+            raise TypeError("A selector's category can only be a Tag.")
+        return selector.clone(
             name=name, category=category, value=value, capture=capture
         )
 
 
-def _select(pattern, context="root"):
-    if isinstance(pattern, str):
-        pattern = parse(pattern)
-    if isinstance(pattern, Element):
-        pattern = Call(
+def _select(selector, context="root"):
+    if isinstance(selector, str):
+        selector = parse(selector)
+    if isinstance(selector, Element):
+        selector = Call(
             element=Element(name=None),
-            captures=(pattern.with_focus(),),
+            captures=(selector.with_focus(),),
             immediate=False,
         )
-    assert isinstance(pattern, Call)
-    return pattern
+    assert isinstance(selector, Call)
+    return selector
 
 
-def select(s, env=None, env_wrapper=None, skip_modules=[], skip_frames=0):
+def select(s, env=None, skip_modules=[], skip_frames=0, strict=False):
     """Create a selector from a string.
 
     Arguments:
@@ -727,13 +826,37 @@ def select(s, env=None, env_wrapper=None, skip_modules=[], skip_frames=0):
             is outside these modules.
         skip_frames: Number of frames to skip when looking for an
             environment.
+        strict: Whether to require functions and variables in the selector
+            to be statically resolvable (will give better errors).
     """
     if not isinstance(s, str):
         return s
     if env is None:
         fr = sys._getframe(skip_frames + 1)
         env = _find_eval_env(s, fr, skip=["ptera", "contextlib", *skip_modules])
-        if env_wrapper is not None:
-            env = env_wrapper(env)
-    pattern = _select(s)
-    return _resolve(pattern, env, count())
+    sel = _select(s)
+    rval = _resolve(sel, env, count())
+
+    if strict:
+        verify(rval, display=s)
+
+    return rval
+
+
+def verify(selector, display=None):
+    """Verify that the selector is resolvable.
+
+    This raises an exception if :func:`~ptera.selector.Call.problems`
+    returns any problems.
+
+    * Wildcards are not allowed for cuntions.
+    * All functions should be tooled.
+    * All captured variables should exist in their respective functions.
+    * For wildcard variables that specify a tag/category, at least one
+      variable should match.
+    """
+    display = display or selector
+    problems = selector.problems()
+    if problems:
+        raise SelectorError(repr(display) + ": " + "\n".join(problems))
+    return selector
